from pathlib import Path

import numpy as np
import pandas as pd
from PIL import Image
from torch.utils.data import Dataset

from histopath.torch.data.openslide_tile_reader import OpenSlideTileReader


class SlideDataset(Dataset[pd.Series], OpenSlideTileReader):
    """Dataset for reading tiles from a single slide image.

    This dataset reads tiles from an OpenSlide image. The tiles are specified by a
    DataFrame with columns ["x", "y"]. The RGBA tiles are converted to RGB before
    being returned.

    Attributes:
        slide (Path): Path to the slide image.
        level (int | str): Level of the slide to read. If int, it is used as the level.
            If str, it is used as the column name in the tiles DataFrame.
        tile_extent_x (int | str): Width of the tile. If int, it is used as the width.
            If str, it is used as the column name in the tiles DataFrame.
        tile_extent_y (int | str): Height of the tile. If int, it is used as the height.
            If str, it is used as the column name in the tiles DataFrame.
        tiles (pd.DataFrame): DataFrame with columns ["x", "y"] specifying the tiles
            to be read.
    """

    def __init__(
        self,
        slide_tiles: pd.DataFrame,
        slide_path: str | Path,
        tile_extent: int | tuple[int, int] | str,
        level: int | str | None,
        slide_resolution: float | tuple[float, float] | None,
        background: None | tuple[int, int, int] = (255, 255, 255),
    ) -> None:
        """Initialize OpenSlideTilesDataset dataset.

        Args:
            slide_path: Path to the slide image.
            level: Level of the slide to read. If int, it is used as the level. If str,
                it is used as the column name in the tiles DataFrame.
            tile_extent_x: Width of the tile. If int, it is used as the width. If str, it
                is used as the column name in the tiles DataFrame.
            tile_extent_y: Height of the tile. If int, it is used as the height. If str,
                it is used as the column name in the tiles DataFrame.
            tiles: DataFrame with columns ["x", "y"].
        """
        super().__init__(
            path=slide_path,
            level=level,
            resolution=slide_resolution,
            background=background,
        )

        self.slide_tiles = slide_tiles
        self.tile_extent = tile_extent

    def __len__(self) -> int:
        return len(self.slide_tiles)

    def __getitem__(self, idx: int) -> pd.Series:
        return self.slide_tiles.iloc[idx]

    def get_tile(self, tile_coords: tuple[int, int], tile: pd.Series) -> Image.Image:
        """Returns tile from the slide image at the specified coordinates in RGB format."""
        tile_extent = self._get_from_tile(tile, self.tile_extent)

        return super().get_openslide_tile(
            tile_coords=tile_coords,
            tile_extent=tuple(np.broadcast_to(tile_extent, 2)),  # type: ignore[call-overload]
            tile=tile,
        )



from typing import Any

import numpy as np
from histopath.openslide import OpenSlide
from PIL import Image


def openslide_tile_reader(row: dict[str, Any]) -> Any:
    with OpenSlide(row["path"]) as slide:
        rgba_region = slide.read_region(
            (row["tile_x"], row["tile_y"]),
            row["level"],
            (row["tile_extent_x"], row["tile_extent_y"]),
        )
        rgb_region = Image.alpha_composite(
            Image.new("RGBA", rgba_region.size, (255, 255, 255)), rgba_region
        ).convert("RGB")
        row["tile"] = np.asarray(rgb_region)

    return row



import hashlib
from typing import Any


def row_hash(row: dict[str, Any], column: str = "id") -> dict[str, Any]:
    """Hashes a row (dictionary) using SHA256 and adds the hash as a new column.

    Args:
        row: The dictionary (row) to hash.
        column: The name of the column to store the hash. Defaults to "id".

    Returns:
        The modified row (dictionary) with the SHA256 hash added.
    """
    row[column] = hashlib.sha256(str(row).encode()).hexdigest()
    return row



from sys import getsizeof
from typing import Iterator

import numpy as np
import pyarrow
from ray.data.block import Block
from ray.data.datasource import FileBasedDatasource

FILE_EXTENSIONS = [
    "svs",
    "tif",
    "dcm",
    "ndpi",
    "vms",
    "vmu",
    "scn",
    "mrxs",
    "tiff",
    "svslide",
    "bif",
    "czi",
]


class OpenSlideMetaDatasource(FileBasedDatasource):
    """Datasource for reading OpenSlide metadata.

    This datasource reads metadata from OpenSlide files and returns a block containing
    the metadata for each file. The metadata includes the slide dimensions, tile extent,
    stride, and resolution (microns per pixel) for the specified level or resolution.

    Args:
        paths: Path(s) to the OpenSlide files.
        mpp: Desired resolution in microns per pixel. If provided, `level` must be None.
        level: Desired level of the slide. If provided, `mpp` must be None.
        tile_extent: Size of the tiles to be generated from the slide.
        stride: Stride for tiling the slide.
        **file_based_datasource_kwargs: Additional arguments for the FileBasedDatasource.

    Raises:
        AssertionError: If both `mpp` and `level` are provided or if neither is provided.

    Example:
        >>> from histopath.ray.datasource.openslide_metadatasource import OpenSlideMetaDatasource
        >>> datasource = OpenSlideMetaDatasource(
        ...     paths=["slide1.svs", "slide2.tiff"],
        ...     mpp=0.25,
        ...     tile_extent=(512, 512),
        ...     stride=(256, 256),
        ... )
    """

    def __init__(
        self,
        paths: str | list[str],
        *,
        mpp: float | None = None,
        level: int | None = None,
        tile_extent: int | tuple[int, int],
        stride: int | tuple[int, int],
        **file_based_datasource_kwargs,
    ) -> None:
        super().__init__(
            paths, file_extensions=FILE_EXTENSIONS, **file_based_datasource_kwargs
        )

        assert (mpp is not None) != (level is not None), (
            "Exactly one of 'mpp' or 'level' must be provided, not both or neither."
        )

        self.desired_mpp = mpp
        self.desired_level = level
        self.tile_extent = np.broadcast_to(tile_extent, 2)
        self.stride = np.broadcast_to(stride, 2)

    def _read_stream(self, f: pyarrow.NativeFile, path: str) -> Iterator[Block]:
        from ray.data._internal.delegating_block_builder import DelegatingBlockBuilder

        from histopath.openslide import OpenSlide

        with OpenSlide(path) as slide:
            if self.desired_level is not None:
                level = self.desired_level
            else:
                assert self.desired_mpp is not None
                level = slide.closest_level(self.desired_mpp)
            mpp_x, mpp_y = slide.slide_resolution(level)

            extent = slide.level_dimensions[level]

        builder = DelegatingBlockBuilder()
        item = {
            "path": path,
            "extent_x": extent[0],
            "extent_y": extent[1],
            "tile_extent_x": self.tile_extent[0],
            "tile_extent_y": self.tile_extent[1],
            "stride_x": self.stride[0],
            "stride_y": self.stride[1],
            "mpp_x": mpp_x,
            "mpp_y": mpp_y,
            "level": level,
        }
        builder.add(item)
        yield builder.build()

    def _rows_per_file(self) -> int:  # type: ignore[override]
        return 1

    def estimate_inmemory_data_size(self) -> int | None:
        paths = self._paths()
        if not paths:
            return 0

        # Create a sample item to calculate the base size of a single row.
        sample_item = {
            "path": "",
            "extent_x": 0,
            "extent_y": 0,
            "tile_extent_x": 0,
            "tile_extent_y": 0,
            "stride_x": 0,
            "stride_y": 0,
            "mpp_x": 0.0,
            "mpp_y": 0.0,
            "level": 0,
        }

        # Calculate the size of the dictionary structure, keys, and fixed-size values.
        base_row_size = getsizeof(sample_item)
        for k, v in sample_item.items():
            base_row_size += getsizeof(k)
            base_row_size += getsizeof(v)

        # Calculate the total size of all path strings.
        total_path_size = sum(getsizeof(p) for p in paths)

        # The total estimated size is the base size for each row plus the total size of paths.
        return base_row_size * len(paths) + total_path_size
